# -*- coding: utf-8 -*-
#
# Copyright 2019 Radosław Piliszek (yoctozepto)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jinja2.filters import pass_context
from jinja2.runtime import Undefined

from kolla_ansible.exception import FilterError
from kolla_ansible.helpers import _call_bool_filter


@pass_context
def kolla_address(context, network_name, hostname=None):
    """returns IP address on the requested network

    The output is affected by '<network_name>_*' variables:
    '<network_name>_interface' sets the interface to obtain address for.
    '<network_name>_address_family' controls the address family (ipv4/ipv6).

    :param context: Jinja2 Context
    :param network_name: string denoting the name of the network to get IP
                         address for, e.g. 'api'
    :param hostname: to override host which address is retrieved for
    :returns: string with IP address
    """

    # NOTE(yoctozepto): watch out as Jinja2 'context' behaves not exactly like
    # the python 'dict' (but mimics it most of the time)
    # for example it returns a special object of type 'Undefined' instead of
    # 'None' or value specified as default for 'get' method
    # 'HostVars' shares this behavior

    if hostname is None:
        hostname = context.get('inventory_hostname')
        if isinstance(hostname, Undefined):
            raise FilterError("'inventory_hostname' variable is unavailable")

    hostvars = context.get('hostvars')
    if isinstance(hostvars, Undefined):
        raise FilterError("'hostvars' variable is unavailable")

    host = hostvars.get(hostname)
    if isinstance(host, Undefined):
        raise FilterError("'{hostname}' not in 'hostvars'"
                          .format(hostname=hostname))

    del hostvars  # remove for clarity (no need for other hosts)

    # NOTE(yoctozepto): variable "host" will *not* return Undefined
    # same applies to all its children (act like plain dictionary)

    interface_name = host.get(network_name + '_interface')
    if interface_name is None:
        raise FilterError("Interface name undefined "
                          "for network '{network_name}' "
                          "(set '{network_name}_interface')"
                          .format(network_name=network_name))

    address_family = host.get(network_name + '_address_family')
    if address_family is None:
        raise FilterError("Address family undefined "
                          "for network '{network_name}' "
                          "(set '{network_name}_address_family')"
                          .format(network_name=network_name))
    address_family = address_family.lower()
    if address_family not in ['ipv4', 'ipv6']:
        raise FilterError("Unknown address family '{address_family}' "
                          "for network '{network_name}'"
                          .format(address_family=address_family,
                                  network_name=network_name))

    ansible_interface_name = interface_name.replace('-', '_')
    interface = host['ansible_facts'].get(ansible_interface_name)
    if interface is None:
        raise FilterError("Interface '{interface_name}' "
                          "not present "
                          "on host '{hostname}'"
                          .format(interface_name=interface_name,
                                  hostname=hostname))

    af_interface = interface.get(address_family)
    if af_interface is None:
        raise FilterError("Address family '{address_family}' undefined "
                          "on interface '{interface_name}' "
                          "for host: '{hostname}'"
                          .format(address_family=address_family,
                                  interface_name=interface_name,
                                  hostname=hostname))

    if address_family == 'ipv4':
        address = af_interface.get('address')
    elif address_family == 'ipv6':
        # ipv6 has no concept of a secondary address
        # explicitly exclude the vip addresses
        # to avoid excluding all /128

        haproxy_enabled = host.get('enable_haproxy')
        if haproxy_enabled is None:
            raise FilterError("'enable_haproxy' variable is unavailable")
        haproxy_enabled = _call_bool_filter(context, haproxy_enabled)

        if haproxy_enabled:
            vip_addresses = [
                host.get('kolla_internal_vip_address'),
                host.get('kolla_external_vip_address'),
            ]
        else:
            # no addresses are virtual (kolla-wise)
            vip_addresses = []

        global_ipv6_addresses = [x for x in af_interface if
                                 x['scope'] == 'global' and
                                 x['address'] not in vip_addresses]

        if global_ipv6_addresses:
            address = global_ipv6_addresses[0]['address']
        else:
            address = None

    if address is None:
        raise FilterError("{address_family} address missing "
                          "on interface '{interface_name}' "
                          "for host '{hostname}'"
                          .format(address_family=address_family,
                                  interface_name=interface_name,
                                  hostname=hostname))

    return address



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from string import Template
from time import sleep

import dbus


TEMPLATE = '''# ${service_name}
# autogenerated by Kolla-Ansible

[Unit]
Description=docker ${service_name}
After=docker.service
Requires=docker.service
StartLimitIntervalSec=${restart_timeout}
StartLimitBurst=${restart_retries}

[Service]
ExecStart=/usr/bin/docker start -a ${name}
ExecStop=/usr/bin/docker stop ${name} -t ${graceful_timeout}
Restart=${restart_policy}
RestartSec=${restart_duration}

[Install]
WantedBy=multi-user.target
'''


class SystemdWorker(object):
    def __init__(self, params):
        name = params.get('name', None)

        # systemd is not needed
        if not name:
            return None

        restart_policy = params.get('restart_policy', 'no')
        if restart_policy == 'unless-stopped':
            restart_policy = 'always'

        # NOTE(hinermar): duration * retries should be less than timeout
        # otherwise service will indefinitely try to restart.
        # Also, correct timeout and retries values should probably be
        # checked at the module level inside kolla_docker.py
        restart_timeout = params.get('client_timeout', 120)
        restart_retries = params.get('restart_retries', 10)
        restart_duration = (restart_timeout // restart_retries) - 1

        # container info
        self.container_dict = dict(
            name=name,
            service_name='kolla-' + name + '-container.service',
            engine='docker',
            deps='docker.service',
            graceful_timeout=params.get('graceful_timeout'),
            restart_policy=restart_policy,
            restart_timeout=restart_timeout,
            restart_retries=restart_retries,
            restart_duration=restart_duration
        )

        # systemd
        self.manager = self.get_manager()
        self.job_mode = 'replace'
        self.sysdir = '/etc/systemd/system/'

        # templating
        self.template = Template(TEMPLATE)

    def get_manager(self):
        sysbus = dbus.SystemBus()
        systemd1 = sysbus.get_object(
            'org.freedesktop.systemd1',
            '/org/freedesktop/systemd1'
        )
        return dbus.Interface(systemd1, 'org.freedesktop.systemd1.Manager')

    def start(self):
        if self.perform_action(
            'StartUnit',
            self.container_dict['service_name'],
            self.job_mode
        ):
            return self.wait_for_unit(self.container_dict['restart_timeout'])
        return False

    def restart(self):
        if self.perform_action(
            'RestartUnit',
            self.container_dict['service_name'],
            self.job_mode
        ):
            return self.wait_for_unit(self.container_dict['restart_timeout'])
        return False

    def stop(self):
        if self.perform_action(
            'StopUnit',
            self.container_dict['service_name'],
            self.job_mode
        ):
            return self.wait_for_unit(
                self.container_dict['restart_timeout'],
                state='dead'
            )
        return False

    def reload(self):
        return self.perform_action(
            'Reload',
            self.container_dict['service_name'],
            self.job_mode
        )

    def enable(self):
        return self.perform_action(
            'EnableUnitFiles',
            [self.container_dict['service_name']],
            False,
            True
        )

    def perform_action(self, function, *args):
        try:
            getattr(self.manager, function)(*args)
            return True
        except Exception:
            return False

    def check_unit_file(self):
        return os.path.isfile(
            self.sysdir + self.container_dict['service_name']
        )

    def check_unit_change(self, new_content=''):
        if not new_content:
            new_content = self.generate_unit_file()

        if self.check_unit_file():
            with open(
                self.sysdir + self.container_dict['service_name'], 'r'
            ) as f:
                curr_content = f.read()

                # return whether there was change in the unit file
                return curr_content != new_content

        return True

    def generate_unit_file(self):
        return self.template.substitute(self.container_dict)

    def create_unit_file(self):
        file_content = self.generate_unit_file()

        if self.check_unit_change(file_content):
            with open(
                self.sysdir + self.container_dict['service_name'], 'w'
            ) as f:
                f.write(file_content)

            self.reload()
            self.enable()
            return True

        return False

    def remove_unit_file(self):
        if self.check_unit_file():
            os.remove(self.sysdir + self.container_dict['service_name'])
            self.reload()

            return True
        else:
            return False

    def get_unit_state(self):
        unit_list = self.manager.ListUnits()

        for service in unit_list:
            if str(service[0]) == self.container_dict['service_name']:
                return str(service[4])

        return None

    def wait_for_unit(self, timeout, state='running'):
        delay = 5
        elapsed = 0

        while True:
            if self.get_unit_state() == state:
                return True
            elif elapsed > timeout:
                return False
            else:
                sleep(delay)
                elapsed += delay



# Copyright (c) 2019 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from unittest import mock

import jinja2

from kolla_ansible import exception
from kolla_ansible import filters

from kolla_ansible.tests.unit.helpers import _to_bool


class TestFilters(unittest.TestCase):

    def setUp(self):
        # Bandit complains about Jinja2 autoescaping without nosec.
        self.env = jinja2.Environment()  # nosec
        self.env.filters['bool'] = _to_bool
        self.context = self._make_context()

    def _make_context(self, parent=None):
        if parent is None:
            parent = {}
        return self.env.context_class(
            self.env, parent=parent, name='dummy', blocks={})

    def test_service_enabled_true(self):
        service = {
            'enabled': True
        }
        self.assertTrue(filters.service_enabled(self.context, service))

    def test_service_enabled_yes(self):
        service = {
            'enabled': 'yes'
        }
        self.assertTrue(filters.service_enabled(self.context, service))

    def test_service_enabled_false(self):
        service = {
            'enabled': False
        }
        self.assertFalse(filters.service_enabled(self.context, service))

    def test_service_enabled_no(self):
        service = {
            'enabled': 'no'
        }
        self.assertFalse(filters.service_enabled(self.context, service))

    def test_service_enabled_no_attr(self):
        service = {}
        self.assertRaises(exception.FilterError,
                          filters.service_enabled, self.context, service)

    def test_service_mapped_to_host_host_in_groups_true(self):
        service = {
            'host_in_groups': True
        }
        self.assertTrue(filters.service_mapped_to_host(self.context, service))

    def test_service_mapped_to_host_host_in_groups_yes(self):
        service = {
            'host_in_groups': 'yes'
        }
        self.assertTrue(filters.service_mapped_to_host(self.context, service))

    def test_service_mapped_to_host_host_in_groups_false(self):
        service = {
            'host_in_groups': False
        }
        self.assertFalse(filters.service_mapped_to_host(self.context, service))

    def test_service_mapped_to_host_host_in_groups_no(self):
        service = {
            'host_in_groups': 'no'
        }
        self.assertFalse(filters.service_mapped_to_host(self.context, service))

    def test_service_mapped_to_host_in_group(self):
        service = {
            'group': 'foo'
        }
        context = self._make_context({'group_names': ['foo', 'bar']})
        self.assertTrue(filters.service_mapped_to_host(context, service))

    def test_service_mapped_to_host_not_in_group(self):
        service = {
            'group': 'foo'
        }
        context = self._make_context({'group_names': ['bar']})
        self.assertFalse(filters.service_mapped_to_host(context, service))

    def test_service_mapped_to_host_no_attr(self):
        service = {}
        self.assertRaises(exception.FilterError,
                          filters.service_mapped_to_host, self.context,
                          service)

    def test_extract_haproxy_services_empty_dict(self):
        example_service = {}
        actual = filters.extract_haproxy_services(
            self.context, example_service)
        # No change
        self.assertDictEqual({}, actual)

    def test_extract_haproxy_services_no_haproxy_dict(self):
        example_service = {
            "keystone-ssh": {
                "container_name": "keystone_ssh",
                "dimensions": {},
                "enabled": True,
                "group": "keystone",
                "healthcheck": {
                    "interval": "30",
                    "retries": "3",
                    "start_period": "5",
                    "test": [
                        "CMD-SHELL",
                        "healthcheck_listen sshd 8023"
                    ],
                    "timeout": "30"
                },
                "image": "keystone-ssh:latest",
                "volumes": [
                    "/etc/kolla/keystone-ssh/:/var/lib/kolla/config_files/:ro",
                    "/etc/localtime:/etc/localtime:ro",
                    "",
                    "kolla_logs:/var/log/kolla/",
                    "keystone_fernet_tokens:/etc/keystone/fernet-keys"
                ]
            }
        }
        actual = filters.extract_haproxy_services(self.context,
                                                  example_service)
        self.assertDictEqual({}, actual)

    def test_extract_haproxy_services_haproxy_dict(self):
        example_service = {
            "keystone": {
                "container_name": "keystone",
                "dimensions": {},
                "enabled": True,
                "group": "keystone",
                "haproxy": {
                    "keystone_admin": {
                        "enabled": True,
                        "external": False,
                        "listen_port": "35357",
                        "mode": "http",
                        "port": "35357",
                        "tls_backend": True
                    },
                    "keystone_external": {
                        "backend_http_extra": [],
                        "enabled": True,
                        "external": True,
                        "listen_port": "5000",
                        "mode": "http",
                        "port": "5000",
                        "tls_backend": True
                    },
                    "keystone_internal": {
                        "backend_http_extra": [],
                        "enabled": True,
                        "external": False,
                        "listen_port": "5000",
                        "mode": "http",
                        "port": "5000",
                        "tls_backend": True
                    }
                },
                "healthcheck": {
                    "interval": "30",
                    "retries": "3",
                    "start_period": "5",
                    "test": [
                        "CMD-SHELL",
                        "healthcheck_curl https://1.2.3.4:5000"
                    ],
                    "timeout": "30"
                },
                "image": "keystone:latest",
                "volumes": [
                    "/etc/kolla/keystone/:/var/lib/kolla/config_files/:ro",
                    "/etc/localtime:/etc/localtime:ro",
                    "",
                    "",
                    "kolla_logs:/var/log/kolla/",
                    "keystone_fernet_tokens:/etc/keystone/fernet-keys"
                ]
            }
        }
        expected = {
            'keystone_admin': {
                'enabled': True,
                'external': False,
                'listen_port': '35357',
                'mode': 'http',
                'port': '35357',
                'tls_backend': True
            },
            'keystone_external': {
                'backend_http_extra': [],
                'enabled': True,
                'external': True,
                'listen_port': '5000',
                'mode': 'http',
                'port': '5000',
                'tls_backend': True
            },
            'keystone_internal': {
                'backend_http_extra': [],
                'enabled': True,
                'external': False,
                'listen_port': '5000',
                'mode': 'http',
                'port': '5000',
                'tls_backend': True
            }
        }
        actual = filters.extract_haproxy_services(self.context,
                                                  example_service)
        self.assertDictEqual(expected, actual)

    def test_extract_two_services_with_haproxy_dict(self):
        example_service = {
            "glance-api": {
                "container_name": "glance_api",
                "dimensions": {},
                "enabled": True,
                "environment": {
                    "http_proxy": "",
                    "https_proxy": "",
                    "no_proxy": "127.0.0.1,localhost,1.2.3.4,1.2.3.4"
                },
                "group": "glance-api",
                "haproxy": {
                    "glance_api": {
                        "backend_http_extra": [
                            "timeout server 6h"
                        ],
                        "custom_member_list": [
                            "server someserver 1.2.3.4:9292 "
                            "check inter 2000 rise 2 fall 5",
                            ""
                        ],
                        "enabled": False,
                        "external": False,
                        "frontend_http_extra": [
                            "timeout client 6h"
                        ],
                        "mode": "http",
                        "port": "9292"
                    },
                    "glance_api_external": {
                        "backend_http_extra": [
                            "timeout server 6h"
                        ],
                        "custom_member_list": [
                            "server someserver 1.2.3.4:9292 "
                            "check inter 2000 rise 2 fall 5",
                            ""
                        ],
                        "enabled": False,
                        "external": True,
                        "frontend_http_extra": [
                            "timeout client 6h"
                        ],
                        "mode": "http",
                        "port": "9292"
                    }
                },
                "healthcheck": {
                    "interval": "30",
                    "retries": "3",
                    "start_period": "5",
                    "test": [
                        "CMD-SHELL",
                        "healthcheck_curl http://localhost:9292"
                    ],
                    "timeout": "30"
                },
                "host_in_groups": True,
                "image": "centos-source-glance-api:latest",
                "privileged": False,
                "volumes": [
                    "/etc/localtime:/etc/localtime:ro",
                    "",
                    "glance:/var/lib/glance/",
                    "",
                    "kolla_logs:/var/log/kolla/",
                    "",
                    ""
                ]
            },
            "glance-tls-proxy": {
                "container_name": "glance_tls_proxy",
                "dimensions": {},
                "enabled": True,
                "group": "glance-api",
                "haproxy": {
                    "glance_tls_proxy": {
                        "backend_http_extra": [
                            "timeout server 6h"
                        ],
                        "custom_member_list": [
                            "server someserver 1.2.3.4:9292 "
                            "check inter 2000 rise 2 fall 5 ssl verify "
                            "required ca-file ca-bundle.trust.crt",
                            ""
                        ],
                        "enabled": True,
                        "external": False,
                        "frontend_http_extra": [
                            "timeout client 6h"
                        ],
                        "mode": "http",
                        "port": "9292",
                        "tls_backend": "yes"
                    },
                    "glance_tls_proxy_external": {
                        "backend_http_extra": [
                            "timeout server 6h"
                        ],
                        "custom_member_list": [
                            "server someserver 1.2.3.4:9292 "
                            "check inter 2000 rise 2 fall 5 ssl verify "
                            "required ca-file ca-bundle.trust.crt",
                            ""
                        ],
                        "enabled": True,
                        "external": True,
                        "frontend_http_extra": [
                            "timeout client 6h"
                        ],
                        "mode": "http",
                        "port": "9292",
                        "tls_backend": "yes"
                    }
                },
                "healthcheck": {
                    "interval": "30",
                    "retries": "3",
                    "start_period": "5",
                    "test": [
                        "CMD-SHELL",
                        "healthcheck_curl -u openstack:asdf 1.2.3.4:9293"
                    ],
                    "timeout": "30"
                },
                "host_in_groups": True,
                "image": "centos-source-haproxy:latest",
                "volumes": [
                    "/etc/localtime:/etc/localtime:ro",
                    "",
                    "kolla_logs:/var/log/kolla/"
                ]
            }
        }
        expected = {
            "glance_api": {
                "backend_http_extra": [
                    "timeout server 6h"
                ],
                'custom_member_list': ['server someserver '
                                       '1.2.3.4:9292 check inter 2000 '
                                       'rise 2 fall 5',
                                       ''],
                "enabled": False,
                "external": False,
                "frontend_http_extra": [
                    "timeout client 6h"
                ],
                "mode": "http",
                "port": "9292"
            },
            "glance_api_external": {
                "backend_http_extra": [
                    "timeout server 6h"
                ],
                'custom_member_list': ['server someserver '
                                       '1.2.3.4:9292 check inter 2000 '
                                       'rise 2 fall 5',
                                       ''],
                "enabled": False,
                "external": True,
                "frontend_http_extra": [
                    "timeout client 6h"
                ],
                "mode": "http",
                "port": "9292"
            },
            "glance_tls_proxy": {
                "backend_http_extra": [
                    "timeout server 6h"
                ],
                'custom_member_list': ['server someserver 1.2.3.4:9292 '
                                       'check inter 2000 rise 2 fall 5 '
                                       'ssl verify required ca-file '
                                       'ca-bundle.trust.crt',
                                       ''],
                "enabled": True,
                "external": False,
                "frontend_http_extra": [
                    "timeout client 6h"
                ],
                "mode": "http",
                "port": "9292",
                "tls_backend": "yes"
            },
            "glance_tls_proxy_external": {
                "backend_http_extra": [
                    "timeout server 6h"
                ],
                'custom_member_list': ['server someserver 1.2.3.4:9292 '
                                       'check inter 2000 rise 2 fall 5 '
                                       'ssl verify required ca-file '
                                       'ca-bundle.trust.crt',
                                       ''],

                "enabled": True,
                "external": True,
                "frontend_http_extra": [
                    "timeout client 6h"
                ],
                "mode": "http",
                "port": "9292",
                "tls_backend": "yes"
            }
        }
        actual = filters.extract_haproxy_services(self.context,
                                                  example_service)
        self.assertDictEqual(expected, actual)

    def test_extract_haproxy_services_haproxy_dict_duplicate(self):
        example_service = {
            "keystone": {
                "container_name": "keystone",
                "dimensions": {},
                "enabled": True,
                "group": "keystone",
                "haproxy": {
                    "keystone_admin": {
                        "enabled": True,
                        "external": False,
                        "listen_port": "35357",
                        "mode": "http",
                        "port": "35357",
                        "tls_backend": True
                    },
                },
                "healthcheck": {
                    "interval": "30",
                    "retries": "3",
                    "start_period": "5",
                    "test": [
                        "CMD-SHELL",
                        "healthcheck_curl https://1.2.3.4:5000"
                    ],
                    "timeout": "30"
                },
                "image": "keystone:latest",
                "volumes": [
                    "/etc/kolla/keystone/:/var/lib/kolla/config_files/:ro",
                    "kolla_logs:/var/log/kolla/"
                ]
            },
            "keystone-ssh": {
                "container_name": "keystone_ssh",
                "dimensions": {},
                "enabled": True,
                "group": "keystone",
                "haproxy": {
                    "keystone_admin": {
                        "enabled": True,
                        "external": False,
                        "listen_port": "35357",
                        "mode": "http",
                        "port": "35357",
                        "tls_backend": True
                    },
                },
                "healthcheck": {
                    "interval": "30",
                    "retries": "3",
                    "start_period": "5",
                    "test": [
                        "CMD-SHELL",
                        "healthcheck_listen sshd 8023"
                    ],
                    "timeout": "30"
                },
                "image": "keystone-ssh:latest",
                "volumes": [
                    "/etc/kolla/keystone-ssh/:/var/lib/kolla/config_files/:ro",
                    "kolla_logs:/var/log/kolla/"
                ]
            }
        }
        self.assertRaises(exception.FilterError,
                          filters.extract_haproxy_services,
                          self.context, example_service)

    @mock.patch.object(filters, 'service_enabled')
    @mock.patch.object(filters, 'service_mapped_to_host')
    def test_service_enabled_and_mapped_to_host(self, mock_mapped,
                                                mock_enabled):
        service = {}
        mock_enabled.return_value = True
        mock_mapped.return_value = True
        self.assertTrue(filters.service_enabled_and_mapped_to_host(
            self.context, service))
        mock_enabled.assert_called_once_with(self.context, service)
        mock_mapped.assert_called_once_with(self.context, service)

    @mock.patch.object(filters, 'service_enabled')
    @mock.patch.object(filters, 'service_mapped_to_host')
    def test_service_enabled_and_mapped_to_host_disabled(self, mock_mapped,
                                                         mock_enabled):
        service = {}
        mock_enabled.return_value = False
        mock_mapped.return_value = True
        self.assertFalse(filters.service_enabled_and_mapped_to_host(
            self.context, service))
        mock_enabled.assert_called_once_with(self.context, service)
        self.assertFalse(mock_mapped.called)

    @mock.patch.object(filters, 'service_enabled')
    @mock.patch.object(filters, 'service_mapped_to_host')
    def test_service_enabled_and_mapped_to_host_not_mapped(self, mock_mapped,
                                                           mock_enabled):
        service = {}
        mock_enabled.return_value = True
        mock_mapped.return_value = False
        self.assertFalse(filters.service_enabled_and_mapped_to_host(
            self.context, service))
        mock_enabled.assert_called_once_with(self.context, service)
        mock_mapped.assert_called_once_with(self.context, service)

    @mock.patch.object(filters, 'service_enabled_and_mapped_to_host')
    def test_select_services_enabled_and_mapped_to_host(self, mock_seamth):
        services = {
            'foo': object(),
            'bar': object(),
            'baz': object(),
        }
        mock_seamth.side_effect = lambda _, s: s != services['bar']
        result = filters.select_services_enabled_and_mapped_to_host(
            self.context, services)
        expected = {
            'foo': services['foo'],
            'baz': services['baz'],
        }
        self.assertEqual(expected, result)



# -*- coding: utf-8 -*-
#
# Copyright 2022 Michal Arbet (kevko)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kolla_ansible.database_shards import database_shards_info


class FilterModule(object):
    """Database shards filters"""

    def filters(self):
        return {
            'database_shards_info': database_shards_info,
        }



# Copyright (c) 2019 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

import jinja2
import os

from kolla_ansible import nova_filters as filters


class TestFilters(unittest.TestCase):

    def _test_extract_cell(self, test_data, cell_name):
        nova_manage_output = {}
        with open(test_data, 'r') as f:
            nova_manage_output['stdout_lines'] = f.readlines()
        return filters.extract_cell(nova_manage_output, cell_name)

    def test_extract_with_no_name(self):
        test_data = os.path.join(os.path.dirname(
            __file__), 'data', 'nova_manage_cli_output_multiple_cells')
        actual = self._test_extract_cell(test_data, '')
        expected = {'cell_name': None,
                    'cell_uuid': '68a3f49e-27ec-422f-9e2e-2a4e5dc8291b',
                    'cell_message_queue': 'rabbit://openstack:94uCcArVNXEJnhhj4JAAAMI432h5j3k2bb3bbjkB@172.28.128.30:5672',  # noqa
                    'cell_database': 'mysql+pymysql://nova:1nhsVDLWjsFSoCsda453bJBsdsbjkabf77sadsdD@172.28.128.252:3305/nova',  # noqa
                    'cell_disabled': 'False'}
        self.assertDictEqual(expected, actual)

    def test_extract_cell0001(self):
        test_data = os.path.join(os.path.dirname(
            __file__), 'data', 'nova_manage_cli_output_multiple_cells')
        actual = self._test_extract_cell(test_data, 'cell0001')
        expected = {'cell_name': 'cell0001',
                    'cell_uuid': '5af50f3f-82ed-47b2-9de8-7dd4eafa7648',
                    'cell_message_queue': 'rabbit://openstack:0flZc3qqwczsisbaT94uCcArVNXEJnhhj4JAAAMI@172.28.128.32:5673/nova_cell0001',  # noqa
                    'cell_database': 'mysql+pymysql://nova:mpyerUgpbQeXjaDV1nhsVDLWjsFSoCss6dHCXK7G@172.28.128.252:3305/nova_cell0001',  # noqa
                    'cell_disabled': 'False'}
        self.assertDictEqual(expected, actual)

    def test_extract_cell0002(self):
        test_data = os.path.join(os.path.dirname(
            __file__), 'data', 'nova_manage_cli_output_multiple_cells')
        actual = self._test_extract_cell(test_data, 'cell0002')
        expected = {'cell_name': 'cell0002',
                    'cell_uuid': '087697bb-bfc2-471d-befb-0c0fcc7630e4',
                    'cell_message_queue': 'rabbit://openstack:d9LaCxGrQX9Lla9aMZcS7fQ5xLH8S5HwHFrl6jdJ@172.28.128.31:5672/nova_cell0002',  # noqa
                    'cell_database': 'mysql+pymysql://nova:Dj3f3l4yv2SuhbsJyv3ahGIwUMa9TKchw6EXtQfp@172.28.128.253:3306/nova_cell0002',  # noqa
                    'cell_disabled': 'True'}
        self.assertDictEqual(expected, actual)

    def test_extract_missing_cell(self):
        test_data = os.path.join(os.path.dirname(
            __file__), 'data', 'nova_manage_cli_output_multiple_cells')
        actual = self._test_extract_cell(test_data, 'cell0003')
        self.assertIsNone(actual)

    def test_extract_duplicate_cell(self):
        test_data = os.path.join(os.path.dirname(
            __file__), 'data', 'nova_manage_cli_output_duplicate_cells')
        self.assertRaisesRegex(jinja2.TemplateRuntimeError, 'duplicates',
                               self._test_extract_cell, test_data, 'cell0001')

    def test_namespace_haproxy_for_cell_with_empty_name(self):
        example_services = {
            'nova-novncproxy': {
                'group': 'some_group',
                'enabled': True,
                'haproxy': {
                    'nova_novncproxy': {
                        'enabled': True,
                        'mode': 'http',
                        'external': False,
                        'port': '1232',
                        'listen_port': '1233',
                        'backend_http_extra': ['timeout tunnel 1h'],
                    },
                    'nova_novncproxy_external': {
                        'enabled': True,
                        'mode': 'http',
                        'external': True,
                        'port': '1234',
                        'listen_port': '1235',
                        'backend_http_extra': ['timeout tunnel 1h'],
                    }
                }
            }
        }
        actual = filters.namespace_haproxy_for_cell(example_services, '')
        # No change
        self.assertDictEqual(example_services, actual)

    def test_namespace_haproxy_for_cell_with_single_proxy(self):
        example_services = {
            'nova-novncproxy': {
                'group': 'some_group',
                'enabled': True,
                'haproxy': {
                    'nova_novncproxy': {
                        'enabled': True,
                        'mode': 'http',
                        'external': False,
                        'port': '1232',
                        'listen_port': '1233',
                        'backend_http_extra': ['timeout tunnel 1h'],
                    },
                    'nova_novncproxy_external': {
                        'enabled': True,
                        'mode': 'http',
                        'external': True,
                        'port': '1234',
                        'listen_port': '1235',
                        'backend_http_extra': ['timeout tunnel 1h'],
                    }
                }
            }
        }
        actual = filters.namespace_haproxy_for_cell(
            example_services, 'cell0001')
        expected = {
            'nova-novncproxy_cell0001': {
                'group': 'some_group',
                'enabled': True,
                'haproxy': {
                    'nova_novncproxy_cell0001': {
                        'enabled': True,
                        'mode': 'http',
                        'external': False,
                        'port': '1232',
                        'listen_port': '1233',
                        'backend_http_extra': ['timeout tunnel 1h'],
                    },
                    'nova_novncproxy_external_cell0001': {
                        'enabled': True,
                        'mode': 'http',
                        'external': True,
                        'port': '1234',
                        'listen_port': '1235',
                        'backend_http_extra': ['timeout tunnel 1h'],
                    }
                }
            }
        }
        self.assertDictEqual(expected, actual)

    def test_namespace_haproxy_for_cell_with_multiple_proxies(self):
        example_services = {
            'nova-novncproxy': {
                'haproxy': {
                    'nova_novncproxy': {},
                    'nova_novncproxy_external': {}
                }
            },
            'nova-spicehtml5proxy': {
                'haproxy': {
                    'nova_spicehtml5proxy': {},
                    'nova_spicehtml5proxy_external': {}
                }
            }
        }
        actual = filters.namespace_haproxy_for_cell(
            example_services, 'cell0002')
        expected = {
            'nova-novncproxy_cell0002': {
                'haproxy': {
                    'nova_novncproxy_cell0002': {},
                    'nova_novncproxy_external_cell0002': {}
                }
            },
            'nova-spicehtml5proxy_cell0002': {
                'haproxy': {
                    'nova_spicehtml5proxy_cell0002': {},
                    'nova_spicehtml5proxy_external_cell0002': {}
                }
            }
        }
        self.assertDictEqual(expected, actual)



# -*- coding: utf-8 -*-
#
# Copyright 2019 Radosław Piliszek (yoctozepto)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kolla_ansible.kolla_address import kolla_address
from kolla_ansible.put_address_in_context import put_address_in_context


class FilterModule(object):
    """IP address filters"""

    def filters(self):
        return {
            'kolla_address': kolla_address,
            'put_address_in_context': put_address_in_context,
        }



#!/usr/bin/env python3

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import stat
import sys

import hvac
import yaml

from kolla_ansible.hashi_vault import hashicorp_vault_client


def readpwd(passwords_file, vault_kv_path, vault_mount_point, vault_namespace,
            vault_addr, vault_role_id, vault_secret_id, vault_token,
            vault_cacert):

    with open(passwords_file, 'r') as f:
        passwords = yaml.safe_load(f.read())

    if os.stat(passwords_file).st_mode & stat.S_IROTH:
        print(f"WARNING: Passwords file \"{passwords_file}\" is"
              " world-readable. The permissions will be changed.")

    if os.stat(passwords_file).st_mode & stat.S_IWOTH:
        print(f"WARNING: Passwords file \"{passwords_file}\" is"
              " world-writeable. The permissions will be changed.")

    if not isinstance(passwords, dict):
        print("ERROR: Passwords file not in expected key/value format")
        sys.exit(1)

    client = hashicorp_vault_client(vault_namespace, vault_addr, vault_role_id,
                                    vault_secret_id, vault_token, vault_cacert)

    vault_kv_passwords = dict()
    for password_key in passwords:
        try:
            password_data = client.secrets.kv.v2.read_secret_version(
                mount_point=vault_mount_point,
                path="{}/{}".format(vault_kv_path, password_key))
        except hvac.exceptions.InvalidPath:
            # Ignore passwords that are not found in Vault
            print("WARNING: '%s' not found in Vault" % password_key)
            vault_kv_passwords[password_key] = None
            continue
        try:
            vault_kv_passwords[password_key] =\
                password_data['data']['data']['password']
        except KeyError:
            vault_kv_passwords[password_key] = password_data['data']['data']

    try:
        os.remove(passwords_file)
    except OSError:
        pass

    flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    mode = 0o640

    with os.fdopen(os.open(passwords_file, flags, mode=mode), 'w') as f:
        yaml.safe_dump(vault_kv_passwords, f)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-p', '--passwords', type=str,
        default=os.path.abspath('/etc/kolla/passwords.yml'),
        help='Path to the passwords.yml file')
    parser.add_argument(
        '-kv', '--vault-mount-point', type=str,
        default='kv',
        help='Path to the KV mount point')
    parser.add_argument(
        '-kvp', '--vault-kv-path', type=str,
        default='kolla_passwords',
        help='Path to store passwords within your configured KV mount point')
    parser.add_argument(
        '-n', '--vault-namespace', type=str,
        default='',
        help='Vault namespace (enterprise only)')
    parser.add_argument(
        '-v', '--vault-addr', type=str,
        required=True,
        help='Address to connect to an existing Hashicorp Vault')
    parser.add_argument(
        '-r', '--vault-role-id', type=str,
        default='',
        help='Role-ID to authenticate to Vault. This must be used in '
             'conjunction with --secret-id')
    parser.add_argument(
        '-s', '--vault-secret-id', type=str,
        default='',
        help='Secret-ID to authenticate to Vault. This must be used in '
             'conjunction with --role-id')
    parser.add_argument(
        '-t', '--vault-token', type=str,
        default='',
        help='Vault token to authenticate to Vault')
    parser.add_argument(
        '-c', '--vault-cacert', type=str,
        default='',
        help='Path to CA certificate file')

    args = parser.parse_args()
    passwords_file = os.path.expanduser(args.passwords)
    vault_kv_path = args.vault_kv_path
    vault_mount_point = args.vault_mount_point
    vault_namespace = args.vault_namespace
    vault_addr = args.vault_addr
    vault_role_id = args.vault_role_id
    vault_secret_id = args.vault_secret_id
    vault_token = args.vault_token
    vault_cacert = os.path.expanduser(args.vault_cacert)

    readpwd(passwords_file, vault_kv_path, vault_mount_point, vault_namespace,
            vault_addr, vault_role_id, vault_secret_id, vault_token,
            vault_cacert)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

import openstackdocstheme


sys.path.insert(0, os.path.abspath('../..'))
# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'openstackdocstheme',
    'sphinx.ext.autodoc',
    'sphinxcontrib.rsvgconverter',
]

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Kolla Ansible'
copyright = '2013, OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = []
html_theme = 'openstackdocs'
# html_static_path = ['static']

# Add any paths that contain "extra" files, such as .htaccess or
# robots.txt.
html_extra_path = ['_extra']

html_theme_options = {
    "show_other_versions": True,
}

# Output file base name for HTML help builder.
htmlhelp_basename = 'kolla-ansibledoc'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index',
     'doc-kolla-ansible.tex',
     'Kolla Ansible Documentation',
     'OpenStack Foundation', 'manual'),
]

# Disable usage of xindy https://bugzilla.redhat.com/show_bug.cgi?id=1643664
latex_use_xindy = False

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/kolla-ansible'
openstackdocs_pdf_link = True
openstackdocs_bug_project = 'kolla-ansible'
openstackdocs_bug_tag = ''
openstack_projects = [
    'bifrost',
    'cinder',
    'cloudkitty',
    'designate',
    'glance',
    'ironic',
    'keystone',
    'kayobe',
    'kolla',
    'kolla-ansible',
    'magnum',
    'manila',
    'networking-sfc',
    'neutron-vpnaas',
    'neutron',
    'nova',
    'octavia',
    'oslo.messaging',
    'oslotest',
    'ovn-octavia-provider',
    'swift',
    'watcher',
]

# Global variables
# For replacement, use in docs as |VAR_NAME| (note there's no space around variable name)
# When adding new variables, make sure you add them to GLOBAL_VARIABLE_MAP dictionary as well

KOLLA_OPENSTACK_RELEASE = openstackdocstheme.ext._get_series_name()

if KOLLA_OPENSTACK_RELEASE == 'latest':
    KOLLA_OPENSTACK_RELEASE = 'master'
    KOLLA_BRANCH_NAME = 'master'
    TESTED_RUNTIMES_GOVERNANCE_URL = 'https://governance.openstack.org/tc/reference/runtimes/'
else:
    KOLLA_BRANCH_NAME = 'stable/{}'.format(KOLLA_OPENSTACK_RELEASE)
    TESTED_RUNTIMES_GOVERNANCE_URL =\
        'https://governance.openstack.org/tc/reference/runtimes/{}.html'.format(KOLLA_OPENSTACK_RELEASE)

GLOBAL_VARIABLE_MAP = {
    '|KOLLA_OPENSTACK_RELEASE|': KOLLA_OPENSTACK_RELEASE,
    '|KOLLA_BRANCH_NAME|': KOLLA_BRANCH_NAME,
    '|KOLLA_BRANCH_NAME_DASHED|': KOLLA_BRANCH_NAME.replace('/', '-'),
    '|TESTED_RUNTIMES_GOVERNANCE_URL|': TESTED_RUNTIMES_GOVERNANCE_URL,
}


def replace_global_vars(app, docname, source):
    # unlike rst_epilog, replaces variables (strings) in code blocks as well
    # thanks to https://github.com/sphinx-doc/sphinx/issues/4054#issuecomment-329097229
    result = source[0]
    for key in app.config.GLOBAL_VARIABLE_MAP:
        result = result.replace(key, app.config.GLOBAL_VARIABLE_MAP[key])
    source[0] = result


def setup(app):
   app.add_config_value('GLOBAL_VARIABLE_MAP', {}, True)
   app.connect('source-read', replace_global_vars)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
import shlex

from ansible.module_utils.kolla_systemd_worker import SystemdWorker

COMPARE_CONFIG_CMD = ['/usr/local/bin/kolla_set_configs', '--check']


class ContainerWorker(ABC):
    def __init__(self, module):
        self.module = module
        self.params = self.module.params
        self.changed = False
        # Use this to store arguments to pass to exit_json().
        self.result = {}
        self._cgroupns_mode_supported = True

        self.systemd = SystemdWorker(self.params)

        # NOTE(mgoddard): The names used by Docker are inconsisent between
        # configuration of a container's resources and the resources in
        # container_info['HostConfig']. This provides a mapping between the
        # two.
        self.dimension_map = {
            'mem_limit': 'Memory', 'mem_reservation': 'MemoryReservation',
            'memswap_limit': 'MemorySwap', 'cpu_period': 'CpuPeriod',
            'cpu_quota': 'CpuQuota', 'cpu_shares': 'CpuShares',
            'cpuset_cpus': 'CpusetCpus', 'cpuset_mems': 'CpusetMems',
            'kernel_memory': 'KernelMemory', 'blkio_weight': 'BlkioWeight',
            'ulimits': 'Ulimits'}

    @abstractmethod
    def check_image(self):
        pass

    @abstractmethod
    def get_container_info(self):
        pass

    @abstractmethod
    def check_container(self):
        pass

    def compare_container(self):
        container = self.check_container()
        if (not container or
                self.check_container_differs() or
                self.compare_config() or
                self.systemd.check_unit_change()):
            self.changed = True
        return self.changed

    def check_container_differs(self):
        container_info = self.get_container_info()
        if not container_info:
            return True

        return (
            self.compare_cap_add(container_info) or
            self.compare_security_opt(container_info) or
            self.compare_image(container_info) or
            self.compare_ipc_mode(container_info) or
            self.compare_labels(container_info) or
            self.compare_privileged(container_info) or
            self.compare_pid_mode(container_info) or
            self.compare_cgroupns_mode(container_info) or
            self.compare_tmpfs(container_info) or
            self.compare_volumes(container_info) or
            self.compare_volumes_from(container_info) or
            self.compare_environment(container_info) or
            self.compare_container_state(container_info) or
            self.compare_dimensions(container_info) or
            self.compare_command(container_info) or
            self.compare_healthcheck(container_info)
        )

    def compare_ipc_mode(self, container_info):
        new_ipc_mode = self.params.get('ipc_mode')
        current_ipc_mode = container_info['HostConfig'].get('IpcMode')
        if not current_ipc_mode:
            current_ipc_mode = None

        # only check IPC mode if it is specified
        if new_ipc_mode is not None and new_ipc_mode != current_ipc_mode:
            return True
        return False

    def compare_cap_add(self, container_info):
        new_cap_add = self.params.get('cap_add', list())
        try:
            current_cap_add = container_info['HostConfig'].get('CapAdd', None)
        except KeyError:
            current_cap_add = None
        except TypeError:
            current_cap_add = None

        if not current_cap_add:
            current_cap_add = list()
        if set(new_cap_add).symmetric_difference(set(current_cap_add)):
            return True

    def compare_security_opt(self, container_info):
        ipc_mode = self.params.get('ipc_mode')
        pid_mode = self.params.get('pid_mode')
        privileged = self.params.get('privileged', False)
        # NOTE(jeffrey4l) security opt is disabled when using host ipc mode or
        # host pid mode or privileged. So no need to compare security opts
        if ipc_mode == 'host' or pid_mode == 'host' or privileged:
            return False
        new_sec_opt = self.params.get('security_opt', list())
        try:
            current_sec_opt = container_info['HostConfig'].get('SecurityOpt',
                                                               list())
        except KeyError:
            current_sec_opt = None
        except TypeError:
            current_sec_opt = None

        if not current_sec_opt:
            current_sec_opt = list()
        if set(new_sec_opt).symmetric_difference(set(current_sec_opt)):
            return True

    @abstractmethod
    def compare_pid_mode(self, container_info):
        pass

    def compare_cgroupns_mode(self, container_info):
        if not self._cgroupns_mode_supported:
            return False
        new_cgroupns_mode = self.params.get('cgroupns_mode')
        if new_cgroupns_mode is None:
            # means we don't care what it is
            return False
        current_cgroupns_mode = (container_info['HostConfig']
                                 .get('CgroupnsMode')) or \
                                (container_info['HostConfig']
                                 .get('CgroupMode'))
        if current_cgroupns_mode in ('', None):
            # means the container was created on Docker pre-20.10
            # it behaves like 'host'
            current_cgroupns_mode = 'host'
        return new_cgroupns_mode != current_cgroupns_mode

    def compare_privileged(self, container_info):
        new_privileged = self.params.get('privileged')
        current_privileged = container_info['HostConfig']['Privileged']
        if new_privileged != current_privileged:
            return True

    @abstractmethod
    def compare_image(self, container_info=None):
        pass

    def compare_labels(self, container_info):
        new_labels = self.params.get('labels')
        current_labels = container_info['Config'].get('Labels', dict())
        image_labels = self.check_image().get('Labels', dict())
        for k, v in image_labels.items():
            if k in new_labels:
                if v != new_labels[k]:
                    return True
            else:
                del current_labels[k]

        if new_labels != current_labels:
            return True

    def compare_tmpfs(self, container_info):
        new_tmpfs = self.generate_tmpfs()
        current_tmpfs = container_info['HostConfig'].get('Tmpfs')
        if not new_tmpfs:
            new_tmpfs = []
        if not current_tmpfs:
            current_tmpfs = []

        if set(current_tmpfs).symmetric_difference(set(new_tmpfs)):
            return True

    def compare_volumes_from(self, container_info):
        new_vols_from = self.params.get('volumes_from')
        current_vols_from = container_info['HostConfig'].get('VolumesFrom')
        if not new_vols_from:
            new_vols_from = list()
        if not current_vols_from:
            current_vols_from = list()

        if set(current_vols_from).symmetric_difference(set(new_vols_from)):
            return True

    @abstractmethod
    def compare_volumes(self, container_info):
        pass

    def compare_dimensions(self, container_info):
        new_dimensions = self.params.get('dimensions')

        if not self._dimensions_kernel_memory_removed:
            self.dimension_map['kernel_memory'] = 'KernelMemory'

        unsupported = set(new_dimensions.keys()) - \
            set(self.dimension_map.keys())
        if unsupported:
            self.module.exit_json(
                failed=True, msg=repr("Unsupported dimensions"),
                unsupported_dimensions=unsupported)
        current_dimensions = container_info['HostConfig']
        for key1, key2 in self.dimension_map.items():
            # NOTE(mgoddard): If a resource has been explicitly requested,
            # check for a match. Otherwise, ensure it is set to the default.
            if key1 in new_dimensions:
                if key1 == 'ulimits':
                    if self.compare_ulimits(new_dimensions[key1],
                                            current_dimensions[key2]):
                        return True
                elif new_dimensions[key1] != current_dimensions[key2]:
                    return True
            elif current_dimensions[key2]:
                # The default values of all currently supported resources are
                # '' or 0 - both falsy.
                return True

    def compare_environment(self, container_info):
        if self.params.get('environment'):
            current_env = dict()
            for kv in container_info['Config'].get('Env', list()):
                k, v = kv.split('=', 1)
                current_env.update({k: v})

            for k, v in self.params.get('environment').items():
                if k not in current_env:
                    return True
                if current_env[k] != v:
                    return True

    def compare_container_state(self, container_info):
        new_state = self.params.get('state')
        current_state = container_info['State'].get('Status')

        if new_state == "started" and current_state == "running":
            return False
        if new_state != current_state:
            return True

    def compare_ulimits(self, new_ulimits, current_ulimits):
        # The new_ulimits is dict, we need make it to a list of Ulimit
        # instance.
        new_ulimits = self.build_ulimits(new_ulimits)

        def key(ulimit):
            return ulimit['Name']

        if current_ulimits is None:
            current_ulimits = []
        return sorted(new_ulimits, key=key) != sorted(current_ulimits, key=key)

    def compare_command(self, container_info):
        new_command = self.params.get('command')
        if new_command is not None:
            new_command_split = shlex.split(new_command)
            new_path = new_command_split[0]
            new_args = new_command_split[1:]
            if (new_path != container_info['Path'] or
                    new_args != container_info['Args']):
                return True

    def compare_healthcheck(self, container_info):
        new_healthcheck = self.parse_healthcheck(
            self.params.get('healthcheck'))
        current_healthcheck = container_info['Config'].get('Healthcheck')

        healthcheck_map = {
            'test': 'Test',
            'retries': 'Retries',
            'interval': 'Interval',
            'start_period': 'StartPeriod',
            'timeout': 'Timeout'}

        if new_healthcheck:
            new_healthcheck = new_healthcheck['healthcheck']
            if current_healthcheck:
                new_healthcheck = dict((healthcheck_map.get(k, k), v)
                                       for (k, v) in new_healthcheck.items())
                return new_healthcheck != current_healthcheck
            else:
                return True
        else:
            if current_healthcheck:
                return True

    def parse_image(self):
        full_image = self.params.get('image')

        if '/' in full_image:
            registry, image = full_image.split('/', 1)
        else:
            image = full_image

        if ':' in image:
            return full_image.rsplit(':', 1)
        else:
            return full_image, 'latest'

    @abstractmethod
    def pull_image(self):
        pass

    @abstractmethod
    def remove_container(self):
        pass

    def generate_tmpfs(self):
        tmpfs = self.params.get('tmpfs')
        if tmpfs:
            # NOTE(mgoddard): Filter out any empty strings.
            tmpfs = [t for t in tmpfs if t]
        return tmpfs

    def generate_volumes(self, binds=None):
        if not binds:
            volumes = self.params.get('volumes') or self.params.get('volume')
        else:
            volumes = binds

        if not volumes:
            return None, None

        vol_list = list()
        vol_dict = dict()

        for vol in volumes:
            if len(vol) == 0:
                continue

            if ':' not in vol:
                vol_list.append(vol)
                continue

            split_vol = vol.split(':')

            if (len(split_vol) == 2 and
               ('/' not in split_vol[0] or '/' in split_vol[1])):
                split_vol.append('rw')

            vol_list.append(split_vol[1])
            vol_dict.update({
                split_vol[0]: {
                    'bind': split_vol[1],
                    'mode': split_vol[2]
                }
            })

        return vol_list, vol_dict

    @abstractmethod
    def build_ulimits(self, ulimits):
        pass

    @abstractmethod
    def create_container(self):
        pass

    @abstractmethod
    def recreate_or_restart_container(self):
        pass

    @abstractmethod
    def start_container(self):
        pass

    def get_container_env(self):
        name = self.params.get('name')
        info = self.get_container_info()
        if not info:
            self.module.fail_json(msg="No such container: {}".format(name))
        else:
            envs = dict()
            for env in info['Config']['Env']:
                if '=' in env:
                    key, value = env.split('=', 1)
                else:
                    key, value = env, ''
                envs[key] = value

            self.module.exit_json(**envs)

    def get_container_state(self):
        name = self.params.get('name')
        info = self.get_container_info()
        if not info:
            self.module.fail_json(msg="No such container: {}".format(name))
        else:
            self.module.exit_json(**info['State'])

    def parse_healthcheck(self, healthcheck):
        if not healthcheck:
            return None

        result = dict(healthcheck={})

        # All supported healthcheck parameters
        supported = set(['test', 'interval', 'timeout', 'start_period',
                         'retries'])
        unsupported = set(healthcheck) - supported
        missing = supported - set(healthcheck)
        duration_options = set(['interval', 'timeout', 'start_period'])

        if unsupported:
            self.module.exit_json(failed=True,
                                  msg=repr("Unsupported healthcheck options"),
                                  unsupported_healthcheck=unsupported)

        if missing:
            self.module.exit_json(failed=True,
                                  msg=repr("Missing healthcheck option"),
                                  missing_healthcheck=missing)

        for key in healthcheck:
            value = healthcheck.get(key)
            if key in duration_options:
                try:
                    result['healthcheck'][key] = int(value) * 1000000000
                except TypeError:
                    raise TypeError(
                        'Cannot parse healthcheck "{0}". '
                        'Expected an integer, got "{1}".'
                        .format(value, type(value).__name__)
                    )
                except ValueError:
                    raise ValueError(
                        'Cannot parse healthcheck "{0}". '
                        'Expected an integer, got "{1}".'
                        .format(value, type(value).__name__)
                    )
            else:
                if key == 'test':
                    # If the user explicitly disables the healthcheck,
                    # return None as the healthcheck object
                    if value in (['NONE'], 'NONE'):
                        return None
                    else:
                        if isinstance(value, (tuple, list)):
                            result['healthcheck'][key] = \
                                [str(e) for e in value]
                        else:
                            result['healthcheck'][key] = \
                                ['CMD-SHELL', str(value)]
                elif key == 'retries':
                    try:
                        result['healthcheck'][key] = int(value)
                    except ValueError:
                        raise ValueError(
                            'Cannot parse healthcheck number of retries.'
                            'Expected an integer, got "{0}".'
                            .format(type(value))
                        )

        return result

    @abstractmethod
    def stop_container(self):
        pass

    @abstractmethod
    def stop_and_remove_container(self):
        pass

    @abstractmethod
    def restart_container(self):
        pass

    @abstractmethod
    def create_volume(self):
        pass

    @abstractmethod
    def remove_volume(self):
        pass

    @abstractmethod
    def remove_image(self):
        pass

    @abstractmethod
    def ensure_image(self):
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2022 Michal Arbet (kevko)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jinja2.filters import pass_context
from jinja2.runtime import Undefined

from kolla_ansible.exception import FilterError
from kolla_ansible.helpers import _call_bool_filter


@pass_context
def database_shards_info(context, hostnames):
    """returns dict with database shards info

       Returned dict looks as example below:

       "database_shards_info": {
           "shards": {
               "0": {
                   "hosts": [
                       "controller0",
                       "controller1",
                       "controller2"
                   ]
               },
               "1": {
                   "hosts": [
                       "controller3",
                       "controller4",
                       "controller5"
                   ]
               }
           },
           "users": [
               {
                   "password": "secret",
                   "shard_id": "0",
                   "user": "root_shard_0"
               },
               {
                   "password": "secret",
                   "shard_id": "0",
                   "user": "backup_shard_0"
               },
               {
                   "password": "secret",
                   "shard_id": "1",
                   "user": "root_shard_1"
               },
               {
                   "password": "secret",
                   "shard_id": "1",
                   "user": "backup_shard_1"
               }
           ]
       }

    :param context: Jinja2 Context
    :param hostnames: List of database hosts
    :returns: Dict with database shards info
    """

    hostvars = context.get('hostvars')
    if isinstance(hostvars, Undefined):
        raise FilterError("'hostvars' variable is unavailable")

    shards_info = {'shards': {}, 'users': []}

    for hostname in hostnames:

        host = hostvars.get(hostname)
        if isinstance(host, Undefined):
            raise FilterError(f"'{hostname}' not in 'hostvars'")

        host_shard_id = host.get('mariadb_shard_id')
        if host_shard_id is None:
            raise FilterError(f"'mariadb_shard_id' is undefined "
                              "for host '{hostname}'")
        else:
            host_shard_id = str(host_shard_id)

        if host_shard_id not in shards_info['shards']:
            shards_info['shards'][host_shard_id] = {'hosts': [hostname]}

            backup_enabled = host.get('enable_mariabackup')
            if backup_enabled is None:
                raise FilterError("'enable_mariabackup' variable is "
                                  "unavailable")
            backup_enabled = _call_bool_filter(context, backup_enabled)

            db_password = host.get('database_password')
            if db_password is None:
                raise FilterError("'database_password' variable is "
                                  "unavailable")

            db_root_prefix = host.get('mariadb_shard_root_user_prefix')
            if db_root_prefix is None:
                raise FilterError("'mariadb_shard_root_user_prefix' variable "
                                  "is unavailable")
            db_user = f"{db_root_prefix}{host_shard_id}"
            user_dict = {'password': db_password, 'user': db_user,
                         'shard_id': host_shard_id}
            shards_info['users'].append(user_dict)

            if backup_enabled:
                db_backup_prefix = host.get('mariadb_shard_backup_user_prefix')
                if db_backup_prefix is None:
                    raise FilterError("'mariadb_shard_backup_user_prefix' "
                                      "variable is unavailable")
                db_user = f"{db_backup_prefix}{host_shard_id}"
                user_dict = {'password': db_password, 'user': db_user,
                             'shard_id': host_shard_id}
                shards_info['users'].append(user_dict)
        else:
            shards_info['shards'][host_shard_id]['hosts'].append(hostname)

    return shards_info
